"""
Utility functions for reading/writing analysis result json files, and comparison.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import datetime
from collections import OrderedDict

import numpy as np

RES_UNIT_STANDARD = {
    "length_unit": "meter",
    "rot_angle_unit": "rad",
    "force_unit": "kN",
    "moment_unit": "kN-m",
}

def read_frame_analysis_result_json(res_file_path, verbose=False):
    """
    
    Parameters
    ----------
    res_file_path : [type]
        [description]
    verbose : bool, optional
        [description], by default False
    
    Returns
    -------
    [type]
        [description]
    """
    with open(res_file_path, 'r') as f:
        json_data = json.loads(f.read())
        if verbose: print('conmech result parsed: {}'.format(res_file_path))

    nD_data = json_data['node_displacement']
    nD = {}
    for nd in nD_data:
        # meter / rad
        nD[nd['node_id']] = nd['displacement']

    fR_data = json_data['fixity_reaction']
    fR = {}
    for fr in fR_data:
        fR[fr['node_id']] = np.array(fr['reaction'])

    eR = {}

    unit_info = {}
    for key_name in RES_UNIT_STANDARD.keys():
        unit_info[key_name] = json_data[key_name] if key_name in json_data else RES_UNIT_STANDARD[key_name]

    return nD, fR, eR, unit_info


def compare_analysis_results(res_A_path, res_B_path):
    pass


"""
Python interface class for the stiffness checker backend.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

import os
import json
from collections import defaultdict, OrderedDict
import numpy as np
from numpy.linalg import norm

from _pystiffness_checker import _StiffnessChecker
from pyconmech.frame_analysis.frame_file_io import read_frame_json, write_frame_json

class StiffnessChecker(object):
    """stiffness checking instance for 3D frame deformation analysis

    Calculating elastic deformation given a 3D frame shape. 
    
    """

    def __init__(self, node_points=None, elements=None, fix_specs=None, material_dicts=None,  
        unit='meter', model_type='frame', model_name=None, verbose=False, checker_engine=None):
        """Init fn for stiffness_checker

        By default, self-weight load is applied. Disable by ``StiffnessChecker.set_self_weight_load(False)``
        
        Parameters
        ----------
        node_points : list of 3-float list, optional
            ``#V x 3`` matrix of vertex coordinates, by default None
        elements : list of 2-int list, optional
            ``#E x 2`` matrix of element end point indices, by default None
        fix_specs : dict, optional
            ``{node_id : fixity_spec}``, where ``fixity_spec`` is a 6-dof 0-1 to specify dof fixed (1) or free(0), by default None
        material_dicts : dict, optional
            #E x material property dictionary, see ``pyconmech.database.material_properties`` for examples, by default None
        unit : str, optional
            unit used in the ``node_points``, by default 'meter'
        model_type : str, optional
            structural model type, can be ``frame`` or ``truss``, by default 'frame'
        model_name : str, optional
            model's name, by default None
        verbose : bool, optional
            verbose output (passed into cpp engine), by default False
        checker_engine : class instance, optional
            ``_StiffnessChecker`` generated from cpp backend, or a numpy-based python solver (not implemented yet), by default None
        
        Raises
        ------
        NotImplementedError
            raise if ``model_type`` is set to ``truss``
        """
        self._model_name = model_name or ''
        if model_type == 'truss':
            raise NotImplementedError('truss model is not supported now...')
        self._model_type = model_type
        #
        self._sc_ins = checker_engine
        self._sc_ins.set_self_weight_load(True)
        #
        self._node_points = node_points or []
        self._elements = elements or []
        self._fix_specs = fix_specs or {}
        if not (len(fix_specs) > 0 and all([len(fix_dofs) == 3 or len(fix_dofs) == 6 for fix_dofs in fix_specs.values()])):
            raise RuntimeError('there needs to be at least one support (fixed) vertex in the model!')
        self._material_dicts = material_dicts
        self.set_nodal_displacement_tol()

    @classmethod
    def from_json(cls, json_file_path=None, verbose=False):
        """init class from a frame json file
        
        Parameters
        ----------
        json_file_path : [type], optional
            [description], by default None
        verbose : bool, optional
            [description], by default False
        
        Returns
        -------
        [type]
            [description]
        """
        assert os.path.exists(json_file_path), "json file not exists!"
        checker_engine = _StiffnessChecker(json_file_path=json_file_path, verbose=verbose)

        node_points, elements, fix_specs, model_type, material_dicts, model_name, unit = \
            read_frame_json(json_file_path, verbose=verbose)

        return cls(node_points=node_points, elements=elements, fix_specs=fix_specs, \
                   model_type=model_type, material_dicts=material_dicts, unit=unit, model_name=model_name, verbose=verbose, \
                   checker_engine=checker_engine)

    @classmethod
    def from_frame_data(cls, node_points, elements, fix_specs, material_dicts, 
        unit='meter', model_type='frame', model_name=None, verbose=False):
        """init class from frame data
        
        Parameters
        ----------
        node_points : list of 3-float list, optional
            ``#V x 3`` matrix of vertex coordinates, by default None
        elements : list of 2-int list, optional
            ``#E x 2`` matrix of element end point indices, by default None
        fix_specs : dict, optional
            ``{node_id : fixity_spec}``, where ``fixity_spec`` is a 6-dof 0-1 to specify dof fixed (1) or free(0), by default None
        material_dicts : dict, optional
            #E x material property dictionary, see ``pyconmech.database.material_properties`` for examples, by default None
        unit : str, optional
            unit used in the ``node_points``, by default 'meter'
        model_type : str, optional
            structural model type, can be ``frame`` or ``truss``, by default 'frame'
        model_name : str, optional
            model's name, by default None
        verbose : bool, optional
            verbose output, passed into cpp backend, by default False
        
        Returns
        -------
        StiffnessChecker
            class instance
        """
        # convert to np array
        fixities = np.array([[int(key)] + fix_dofs for key, fix_dofs in fix_specs.items()], dtype=np.int32)
        checker_engine = _StiffnessChecker(np.array(node_points, dtype=np.float64), np.array(elements, dtype=np.int32), fixities, \
            material_dicts, verbose=verbose)
        checker_engine.set_self_weight_load(True)

        return cls(node_points=node_points, elements=elements, fix_specs=fix_specs, \
                   model_type=model_type, material_dicts=material_dicts, unit=unit, model_name=model_name, verbose=verbose, \
                   checker_engine=checker_engine)

    # ==========================================================================
    # properties
    # ==========================================================================

    @property
    def model_name(self):
        return self._model_name

    @property
    def node_points(self):
        """
        
        Returns
        -------
        list of 3-float list
            ``#V x 3`` matrix of vertex coordinates, by default None
        """
        return self._node_points
    
    @property
    def elements(self):
        """

        Returns
        -------
        list of 2-int list
            ``#E x 2`` matrix of element end point indices, by default None
        """
        return self._elements
    
    @property
    def fix_node_ids(self):
        """fixed node indices
        
        Returns
        -------
        list of int
        """
        return list(self._fix_specs.keys())

    @property
    def fix_element_ids(self):
        """return grounded element ids
        
        Returns
        -------
        list of int
            indices of elements with one or more end points fixed
        """
        fix_e_ids = []
        for e_id, e in enumerate(self.elements):
            if any([v_id in self.fix_node_ids for v_id in e]):
                fix_e_ids.append(e_id)
        return fix_e_ids

    @property
    def fix_specs(self):
        """
        
        Returns
        -------
        dict
            ``{node_id : fixity_spec}``, where ``fixity_spec`` is a 6-dof 0-1 to specify dof fixed (1) or free(0), by default None
        """
        return self._fix_specs

    @property
    def model_type(self):
        return self._model_type

    @property
    def materials(self):
        """
        
        Returns
        -------
        dict
            #E x material property dictionary, see ``pyconmech.database.material_properties`` for examples, by default None
        """
        return self._material_dicts

    # ==========================================================================
    # solve functions
    # ==========================================================================

    def solve(self, exist_element_ids=[], if_cond_num=False, eid_sanity_check=False):
        """compute elastic deformation using the linear stiffness equation, return
        True or False indicating if the set criteria is passed nor not.
        
        Parameters
        ----------
        exist_element_ids : list, optional
            list of existing element ids, by default []
            By default (exist_element_ids = []), the stiffness checker will 
            compute for the full structure.
        if_cond_num : bool, optional
            check condition number of the stiffness matrix, by default False
            This will increase the fidelity of the computation but induce some overhead.
            Condition number checking is currently not implemented.
        eid_sanity_check : bool, optional
            check if the given element ids are within range. This is simply a sanity check
            but might induce some extra overhead if the solve function is called many times.
            by default False.
        
        Returns
        -------
        bool
            success or not

        """
        if eid_sanity_check:
            e_id_range = list(range(len(self.elements)))
            for e_id in exist_element_ids:
                assert e_id in e_id_range, 'element id not within range!'
        return self._sc_ins.solve(exist_element_ids, if_cond_num)

    # ==========================================================================
    # load settings
    # ==========================================================================

    def set_loads(self, point_loads=None, include_self_weight=False, gravity_direction=[0,0,-1.0], uniform_distributed_load={}):
        """set load case for the stiffness checker.
        
        Parameters
        ----------
        point_loads : dict, optional
            {node_id : [Fx, Fy, Fz, Mxx, Myy, Mzz]}, in global coordinate,
            by default {}
        include_self_weight : bool, optional
            include gravity load or not, by default False
        uniform_distributed_load : dict, optional
            elemental uniformly distributed load, by default {}
            {element_id : [wx, wy, wz]}, in global cooridinate
        """
        self._sc_ins.set_self_weight_load(include_self_weight=include_self_weight, gravity_direction=gravity_direction)
        if point_loads:
            pt_loads = []
            for vid, vload in point_loads.items():
                pt_loads.append([vid] + vload)
            self._sc_ins.set_load(np.array(pt_loads))
        if uniform_distributed_load:
            ud_loads = []
            for eid, eload in uniform_distributed_load.items():
                ud_loads.append([eid] + eload)
            self._sc_ins.set_uniformly_distributed_loads(np.array(ud_loads))

    def set_self_weight_load(self, include_self_weight, gravity_direction=[0,0,-1.0]):
        """Turn on/off self-weight load.
        
        Parameters
        ----------
        include_self_weight : bool
        """
        self._sc_ins.set_self_weight_load(include_self_weight, gravity_direction=gravity_direction)

    # ==========================================================================
    # check criteria settings
    # ==========================================================================

    def set_nodal_displacement_tol(self, trans_tol=1e-3, rot_tol=0.1745):
        """Set nodal displacement tolerance for stiffness checking criteria.

        If the maximal nodal displacement after the stiffness solve exceeds the tolerance,
        then stiffness_checker.solve() (or with input [ids]) will return False.
        
        Parameters
        ----------
        trans_tol : float, optional
            maximal allowable nodal translational movement in global coordinate, unit in meter, by default 1e-3
        rot_tol : float, optional
            maximal allowable nodal rotational movement in global coordinate, unit in meter, by default 0.1745 
            (about 10 degrees)
        """
        self._sc_ins.set_nodal_displacement_tol(trans_tol, rot_tol)

    def get_nodal_deformation_tol(self):
        """Get nodal displacement tolerance for stiffness checking criteria.
        
        Returns
        -------
        trans_tol : float
            maximal nodal translational deformation, entry-wise
        rot_tol : float
            maximal nodal rotational deformation, entry-wise
        """
        trans_tol, rot_tol = self._sc_ins.get_nodal_deformation_tol()
        return trans_tol, rot_tol

    # TODO: pass in user-specified criteria checking function handle

    # ==========================================================================
    # stiffness data query
    # ==========================================================================

    def has_stored_result(self):
        """Check if the stiffness checker has solved results stored,
        i.e. solve function has been called at least once

        Returns
        -------
        bool
        """
        return self._sc_ins.has_stored_result()
    
    def get_solved_results(self, existing_ids=[]):
        """Fetch back solved results from last time.
        
        Returns
        -------
        success : bool
            pass criteria or not
        nD : dict
            nodal displacements in the global coordinate
            {node_id: np.array([dx, dy, dz, rxx, ryy, rzz]), ...]
        fR : dict
            fixities reaction force and moment in the global coordinate
            {node_id: np.array([Fx, Fy, Fz, Mxx, Myy, Mzz], ...}
        eR : numpy array
            elemental reaction force and moment in the local coordinate
            {element_id: {0 : np.array([F_0_lx, F_0_ly, F_0_lz, M_0_lxx, M_0_lyy, M_0_lzz]),
                          1 : np.array([F_L_lx, F_L_ly, F_L_lz, M_L_lxx, M_L_lyy, M_L_lzz])
            ], ...]
            F_0_lx means internal reaction force at the end point 0, in the direction of local x axis
            M_L_lyy means internal reaction moment at the end point L, around the local yy axis            

            The elemental local coordinate is placed at the end point 0, with local axis pointing along
            the `end point 0` -> `end point L` direction. See doc (TODO: link) for more info.
        """
        if not existing_ids:
            existing_e_ids = list(range(len(self.elements)))
        existing_n_ids = self.get_element_connected_node_ids(existing_ids=existing_ids)
        success, nD_np, fR_np, eR_np = self._sc_ins.get_solved_results()
        nD = {}
        for row in nD_np:
            if int(row[0]) in existing_n_ids:
                nD[int(row[0])] = np.array(row[1:7])
        fR = {}
        for row in fR_np:
            if int(row[0]) in existing_n_ids and int(row[0]) in self.fix_node_ids:
                fR[int(row[0])] = np.array(row[1:7])
        eR = {}
        for row in eR_np:
            if int(row[0]) in existing_e_ids:
                e_r = {}
                e_r[0] = np.array(row[1:7])
                e_r[1] = np.array(row[7:13])
                eR[int(row[0])] = e_r
        return success, nD, fR, eR

    def get_max_nodal_deformation(self):
        """Get max nodal deformation info
        
        Returns
        -------
        max_trans : float
            maximal translational deformation, componentwise max, unit meter
        max_rot : float
            maximal rotational deformation, componentwise max, unit rad
        max_trans_vid : int
            node id for maximal trans deformation
        max_rot_vid : int
            node id for maximal rotational deformation
        """
        max_trans, max_rot, max_trans_vid, max_rot_vid = self._sc_ins.get_max_nodal_deformation()
        return max_trans, max_rot, max_trans_vid, max_rot_vid

    def get_compliance(self):
        """Get compliance of the last solved deformation

        compliance = nodal_loads.dot(nodal_deformation)  
        TODO: check if we need moment and rotational deformation here?
        
        Returns
        -------
        float
        """
        return self._sc_ins.get_compliance()

    def get_self_weight_loads(self, existing_ids=[], dof_flattened=False):
        """Return lumped gravity loads

        TODO: put a link to doc here
        
        Parameters
        ----------
        existing_ids : list, optional
            existing element's ids in the partial structure, by default [], which means full structure
        dof_flattened : bool, optional
            if True, return a flattened dof x 1 vector, 
            otherwise return a dict{node_id: np.array([1:dof])}, by default False
        
        Returns
        -------
        dict (dof_flattened = False)
            {node_id : [Fx, Fy, Fz, Mxx, Myy, Mzz]} in global coordinate, only nodes that exist in the partial structure
            will be returned
        or np.array (dof_flattened = True)
            [Fx, Fy, Fz, Mxx, Myy, Mzz, ...], all nodal dofs will be returned, including nodes that do not
            exist in the partial structure specified in ``existing_ids``
        """
        assert self.model_type == 'frame', 'this function assumes 6 dof each node for now!'
        nL_flat = self._sc_ins.get_gravity_nodal_loads(existing_ids)
        existing_node_ids = self.get_element_connected_node_ids(existing_ids)
        node2dof_map = self.get_node2dof_id_map()
        if not dof_flattened:
            nL = {}
            for nid in existing_node_ids:
                nL[nid] = nL_flat[node2dof_map[nid]]
            return nL
        else:
            return nL_flat


    def get_nodal_loads(self, existing_ids=[], dof_flattened=False):
        """Return nodal loads
        
        Parameters
        ----------
        existing_ids : list, optional
            existing element's ids in the partial structure, by default [], which means full structure
        dof_flattened : bool, optional
            if True, return a flattened dof x 1 vector, 
            otherwise return a dict{node_id: np.array([1:dof])}, by default False
        
        Returns
        -------
        dict (dof_flattened = False)
            {node_id : [Fx, Fy, Fz, Mxx, Myy, Mzz]} in global coordinate, only nodes that exist in the partial structure
            will be returned
        or np.array (dof_flattened = True)
            [Fx, Fy, Fz, Mxx, Myy, Mzz, ...], all nodal dofs will be returned, including nodes that do not
            exist in the partial structure specified in ``existing_ids``
        """
        assert self.model_type == 'frame', 'this function assumes 6 dof each node for now!'
        nL_flat = self._sc_ins.get_lumped_nodal_loads(existing_ids)
        existing_node_ids = self.get_element_connected_node_ids(existing_ids)
        node2dof_map = self.get_node2dof_id_map()
        if not dof_flattened:
            nL = {}
            for nid in existing_node_ids:
                nL[nid] = nL_flat[node2dof_map[nid]]
            return nL
        else:
            return nL_flat

    def get_element_stiffness_matrices(self, in_local_coordinate=False):
        """Get all elemental stiffness matrices, each of which is 12 x 12
        in global coordinate (default): R_{LG}^T * K_{eL} * R_{LG}
        in local coordinate (default): K_{eL}

        Parameters
        ----------
        in_local_coordinate : bool, optional
            return stiffness matrix in the element's local coordinate or not, by default False
        
        Returns
        -------
        dict
            {element_id : np.array}
        """
        eMs = {}
        eMs_raw = self._sc_ins.get_element_stiffness_matrices()
        if in_local_coordinate:
            R_lgs = self.get_element_local2global_rot_matrices()

        for eid, eM_G in enumerate(eMs_raw):
            if not in_local_coordinate:
                eMs[eid] = eM_G
            else:
                R_LG = R_lgs[eid]
                eM_L = R_LG.dot(eM_G)
                eM_L = eM_L.dot(np.transpose(R_LG))
                eMs[eid] = eM_L
        return eMs

    def get_element_local2global_rot_matrices(self):
        """Get all elemental local to global transformation matrices, each of which is 12 x 12

        The array is in the following shape:
        | R       |
        |   R     |
        |     R   |
        |       R |
        where R is the 3x3 coordinate transformation matrix, which tranforms 
        the global xyz axis to the element local axis.

        The coordinate transformation matrix can be used to:
         - transform frame element end forces from the element (local) coordinate system
           to the structure (global) coordinate system
         - transfrom end displacements from the structural (global) coordinate system 
           to the element (local) coordinate system,
         - transform the frame element stiffness and mass matrices
           from element (local) coordinates to structral (global) coordinates.
        Symbolically, the return matrix R = {local}_R_{global}

        Returns
        -------
        dict
            {element_id : 12 x 12 np.array}

        """
        return {nid : mat for nid, mat in enumerate(self._sc_ins.get_element_local2global_rot_matrices())}

    def get_element2dof_id_map(self):
        """Get element_id-to-dof_id maps
        
        Returns
        -------
        dict
            {e_id : {0 : [dof_ids for end point 0]}, {1 : [dof_ids for end point 1]}}
        """
        assert self.model_type == 'frame', 'this function assumes 6 dof each node for now!'
        return {int(e_id) : {0 : id_map[0:6], 1 : id_map[6:12]} 
                for e_id, id_map in enumerate(self._sc_ins.get_element2dof_id_map())}

    def get_node2dof_id_map(self):
        """Get node_id-to-dof_id maps
        
        Returns
        -------
        dict
            {node_id : [dof ids]}
        """
        return {int(n_id) : id_map for n_id, id_map in enumerate(self._sc_ins.get_node2dof_id_map())}

    # ==========================================================================
    # output settings
    # ==========================================================================

    def set_output_json(self, do_output=True, output_dir=None, file_name=None):
        """DEPRECATED

        """
        self._sc_ins.set_output_json(do_output)
        if do_output:
            if not output_dir:
                output_dir = os.path.dirname(os.path.abspath(__file__))
                print('NOTE: no output path is specified, the result file is saved at:\n {}'.format(output_dir))
            else:
                if output_dir[-1] != os.sep and output_dir[-2:] != os.sep:
                    output_dir += os.sep # path sep required by the cpp extmodule
            if not file_name:
                file_name = self.model_name + '_result' + '.json'
            self._sc_ins.set_output_json_path(output_dir, file_name)

    @property
    def result_data(self):
        return self.get_result_data(False)

    def get_result_data(self, output_frame_transf=False):
        if not self.has_stored_result():
            print('no result to output!')
            return None

        success, nD, fR, eR = self.get_solved_results()
        trans_tol, rot_tol = self.get_nodal_deformation_tol()
        max_trans, max_rot, max_t_id, max_r_id = self.get_max_nodal_deformation()
        eR_LG_mats = self.get_element_local2global_rot_matrices()

        data = OrderedDict()
        data['model_name'] = self.model_name
        data['model_type'] = self.model_type
        data['solve_success'] = success
        data['trans_tol'] = trans_tol
        data['rot_tol'] = rot_tol

        data['length_unit'] = 'meter'
        data['angle_unit'] = 'rad'
        data["force_unit"] =  "kN"
        data["moment_unit"] = "kN-m"
        data['compliance'] = self.get_compliance()
        data['max_trans'] = {'node_id' : max_t_id, 'value': max_trans}
        data['max_rot'] = {'node_id' : max_r_id, 'value': max_rot}
    
        nD_data = []
        for n_id, nd in nD.items():
            nd_data = OrderedDict()
            nd_data['node_id'] = n_id
            nd_data['node_pose'] = list(self.node_points[n_id])
            nd_data['displacement'] = nd.tolist()
            nD_data.append(nd_data)
        data['node_displacement'] = nD_data

        eR_data = []
        for e_id, er in eR.items():
            er_data = OrderedDict()
            er_data['element_id'] = e_id
            er_data['node_ids'] = self.elements[e_id]
            er_data['reaction'] = OrderedDict()
            er_data['reaction'][0] = er[0].tolist()
            er_data['reaction'][1] = er[1].tolist()
            if output_frame_transf:
                eR33 = eR_LG_mats[e_id][:3, :3]
                er_data['local_to_global_transformation'] = eR33.tolist()

            eR_data.append(er_data)
        data['element_reaction'] = eR_data

        fR_data = []
        for v_id, fr in fR.items():
            fr_data = OrderedDict()
            fr_data['node_id'] = v_id
            fr_data['node_pose'] = list(self.node_points[v_id])
            fr_data['reaction'] = fr.tolist()
            fR_data.append(fr_data)
        data['fixity_reaction'] = fR_data
        return data

    def write_result_to_json(self, file_path, formatted_json=False):
        with open(file_path, 'w+') as outfile:
            if formatted_json:
                json.dump(self.result_data, outfile, indent=4)
            else:
                json.dump(self.result_data, outfile)

    # ==========================================================================
    # frame data query (TODO: moved to frame class)
    # ==========================================================================

    def get_element_local_node_id(self, element_id, point):
        """Return a point's local id in an element's local coordinate, either 0 or 1
        
        Parameters
        ----------
        element_id : int`
        point : 3-list of float
        """
        end_node_ids = self.elements[element_id]
        if norm(point, self.node_points[end_node_ids[0]]):
            return 0
        elif norm(point, self.node_points[end_node_ids[1]]):
            return 1
        else:
            return None

    def get_node_neighbors(self, return_e_id=False):
        """Return all nodes' connected element ids
        
        Parameters
        ----------
        return_e_ids : bool, optional
            return element ids or the original tuples, by default False

        Returns
        -------
        node_neighbors : dict
            {node_id : set(connected element tuples (n1, n2))}
            or
            {node_id : set(connected element ids)}
        """
        node_neighbors = defaultdict(set)
        for e in self.elements:
            n1, n2 = e
            if not return_e_id:
                node_neighbors[n1].add(e)
                node_neighbors[n2].add(e)
            else:
                e_id = self.elements.index(e)
                node_neighbors[n1].add(e_id)
                node_neighbors[n2].add(e_id)
        return node_neighbors
    
    def get_element_neighbors(self):
        """Return all elements' connected element ids
        
        Returns
        -------
        element_neighbors : dict
            {element_id : set(connected element_ids)}
        """
        node_neighbors = self.get_node_neighbors()
        element_neighbors = defaultdict(set)
        for e in self.elements:
            n1, n2 = e
            element_neighbors[e].update(node_neighbors[n1])
            element_neighbors[e].update(node_neighbors[n2])
            element_neighbors[e].remove(e)
        return element_neighbors

    def get_element_connected_node_ids(self, existing_ids=[], fix_node_only=False):
        """return connected nodes' indices of the elements specified in ``existing_ids``
        
        Parameters
        ----------
        existing_ids : list of int, optional
            element indices, by default [] (full structure)
        fix_node_only : bool, optional
            only return the nodes that are fixed, by default False
        
        Returns
        -------
        list
        """
        if not existing_ids:
            existing_ids = list(range(len(self.elements)))
        connected_node_ids = set()
        for e_id in existing_ids:
            if not fix_node_only:
                connected_node_ids.update([v_id for v_id in self.elements[e_id]])
            else:
                connected_node_ids.update([v_id for v_id in self.elements[e_id] if v_id in self.fix_node_ids])
        return list(connected_node_ids)

    # ==========================================================================
    # beam shape query for deformation visualization
    # ==========================================================================

    def get_original_shape(self, disc=1, draw_full_shape=True):
        pass

    def get_deformed_shape(self, disc=10, exagg_ratio=1.0):
        pass


from pyconmech.frame_analysis import StiffnessChecker

def stiffness_checker_parse_and_solve(file_path, existing_ids=[], pt_loads=None, include_sw=False, result_file_path=None):
    """Wrapper function call to create a StiffnessChecker, solve, and return results
    Mainly for the rpc call, since it doesn't support class creation.
    
    Parameters
    ----------
    file_path : [type]
        [description]
    existing_ids : list, optional
        [description], by default []
    pt_loads : [type], optional
        [description], by default None
    include_sw : bool, optional
        [description], by default False
    
    Returns
    -------
    [type]
        [description]
    """
    sc = StiffnessChecker(json_file_path=file_path)
    sc.set_loads(point_loads=pt_loads, include_self_weight=include_sw)
    sc.solve(existing_ids, eid_sanity_check=True)

    if result_file_path:
        sc.write_result_to_json(result_file_path)
    else:
        return sc.result_data


""" CMake util functions for building python extension module with pybind11 & CMake build system

this script is adapted from cmake_example project for pybind11
https://github.com/pybind/cmake_example

for more info for extension modules:
https://docs.python.org/2/distutils/setupscript.html
"""
from __future__ import absolute_import, print_function

import os
import sys
import re
import platform
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from distutils.version import LooseVersion


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        # execdir = sys.exec_prefix
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable,
                    #   '-DPYTHON_LIBRARY=' + os.path.join(execdir, 'lib', ),
                    #   '-DPYTHON_INCLUDE_DIR=' + ,
                      # disable cpp test because we exclude them in MANIFEST.in
                      '-DCONMECH_BUILD_TESTS=OFF']

        # build against the non-system (homebrew, etc.) Python install
        # https://github.com/pybind/pybind11/issues/99#issuecomment-207782425
        # cmake -DPYTHON_LIBRARY=$(python-config --prefix)/lib/libpython2.7.dylib 
        #       -DPYTHON_INCLUDE_DIR=$(python-config --prefix)/include/python2.7 .

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
            build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


"""
Utility functions for reading/writing frame json files.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import json
import datetime
from collections import OrderedDict

'''length scale conversion to meter'''
LENGTH_SCALE_CONVERSION = {
    'millimeter': 1e-3,
    'meter': 1.0,
}

def parse_point(json_point, scale=1.0, dim=3):
    if dim == 3:
        return [scale*json_point['X'], scale*json_point['Y'], scale*json_point['Z']]
    else:
        return [scale*json_point['X'], scale*json_point['Y']]


def parse_elements(json_data):
    return [tuple(json_element['end_node_ids']) for json_element in json_data['element_list']]


def parse_node_points(json_data, scale=1.0, dim=3):
    return [parse_point(json_node['point'], scale=scale, dim=dim) for json_node in json_data['node_list']]


def parse_fixed_nodes(json_data, dim=3):
    """[summary]
    
    Parameters
    ----------
    json_data : [type]
        [description]
    dim : int, optional
        [description], by default 3
    
    Returns
    -------
    fixities_node_id
        list of int
    fixities_spec
        list of [0-1 specs]
    """
    fixities_spec = {}
    for i, json_node in enumerate(json_data['node_list']):
        if json_node['is_grounded'] == 1:
            if 'fixities' in json_node and json_node['fixities']:
                if dim == 3:
                    assert len(json_node['fixities']) == dim*2 
                else:
                    assert len(json_node['fixities']) == 3 and dim == 2
                fixities_spec[i] = json_node['fixities']
            else:
                if dim == 3:
                    fixities_spec[i] = [1] * 6
                else:
                    fixities_spec[i] = [1] * 3
    return fixities_spec


def check_material_dict(mat_dict):
    """check if a material dict is feasible

    example:
    {
    "material_name": "Steel-S235",
    "youngs_modulus": 21000.0,
    "youngs_modulus_unit": "kN/cm2",
    "density": 78.5,
    "density_unit": "kN/m3",
    "poisson_ratio": 0.300149,
    "cross_sec_area": 19.634954084936208,
    "cross_sec_area_unit": "centimeter^2",
    "_cross_sec_area_note": "round_shape, radius 2.5 cm",
    "Jx": 61.35923151542565,
    "Jx_unit": "centimeter^4",
    "_Jx_note": "0.5 * pi * r^4",
    "Iy": 30.679615757712824,
    "Iy_unit": "centimeter^4",
    "_Iy_note": "0.25 * pi * r^4",
    "Iz": 30.679615757712824,
    "Iz_unit": "centimeter^4",
    "_Iz_note": "0.25 * pi * r^4"}

    Parameters
    ----------
    mat_dict : dict
    
    Returns
    -------
    bool
        valid or not
    ------
    """
    valid = "youngs_modulus" in mat_dict and \
            "youngs_modulus_unit" in mat_dict and \
            "density" in mat_dict and \
            "density_unit" in mat_dict and \
            "poisson_ratio" in mat_dict and \
            "cross_sec_area" in mat_dict and \
            "Jx" in mat_dict and \
            "Iy" in mat_dict and \
            "Iz" in mat_dict

    # TODO: do other unit conversions
    unit_valid = mat_dict["youngs_modulus_unit"] == 'kN/cm2' and \
                 mat_dict["density_unit"] == 'kN/m3' and \
                 mat_dict["cross_sec_area_unit"] == 'centimeter^2' and \
                 mat_dict["Jx_unit"] == 'centimeter^4' and \
                 mat_dict["Iy_unit"] == 'centimeter^4' and \
                 mat_dict["Iz_unit"] == 'centimeter^4'
    return valid and unit_valid


def extract_model_name_from_path(file_path):
    forename = file_path.split('.json')[0]
    return forename.split(os.sep)[-1]


def read_frame_json(file_path, verbose=False):
    """[summary]
    
    Parameters
    ----------
    file_path : [type]
        [description]
    verbose : bool, optional
        [description], by default False
    
    Returns
    -------
    node_points
        list of [x,y,z]
    element_vids, 
        list of [v_id, v_id]
    fix_node_ids
    
    fix_specs, model_type, material_dicts, model_name
    
    Raises
    ------
    ValueError
        [description]
    """
    assert os.path.exists(file_path) and "json file path does not exist!"
    with open(file_path, 'r') as f:
        json_data = json.loads(f.read())
    
    dim = json_data['dimension']
    unit = json_data['unit']
    scale = LENGTH_SCALE_CONVERSION[unit]
    model_type = json_data['model_type']
    model_name = json_data['model_name'] if 'model_name' in json_data else extract_model_name_from_path(file_path)

    if 'frame' in model_type:
        model_type = 'frame'
    elif 'truss' in model_type:
        model_type = 'truss'
    else:
        raise ValueError('model type not supported! Must be frame or truss.')

    element_vids = parse_elements(json_data)
    node_points = parse_node_points(json_data, scale=scale, dim=dim)
    fix_specs = parse_fixed_nodes(json_data, dim)

    if 'uniform_cross_section' in json_data and \
       'uniform_material_properties' in json_data and \
       'material_properties' in json_data and \
       json_data["uniform_cross_section"] and \
       json_data["uniform_material_properties"]:
        material_dicts = [json_data['material_properties']] * len(element_vids)
    else:
        material_dicts = [json_element['material_properties'] for json_element in json_data['element_list']]

    if 'node_num' in json_data:
        num_node = json_data['node_num']
        assert len(node_points) == num_node
    if 'element_num' in json_data:
        num_element = json_data['element_num']
        assert len(element_vids) == num_element
    node_id_range = list(range(len(node_points)))
    for ev1, ev2 in element_vids:
        assert ev1 in node_id_range and ev2 in node_id_range, 'element end point id not in node_list id range!'

    if verbose:
        print('Model: {} | Unit: {}'.format(json_data['model_type'], json_data['unit']))
        print('Nodes: {} | Ground: {} | Elements: {}'.format(
            len(node_points), len(fix_specs), len(element_vids)))
   
    return node_points, element_vids, fix_specs, model_type, material_dicts, model_name, unit


def write_frame_json(file_path, nodes, elements, fixity_specs, material_dicts,
    unif_cross_sec=False, unif_material=False, unit=None, model_type='frame', model_name=None, indent=None, check_material=True):
    data = OrderedDict()
    data['model_name'] = model_name if model_name else extract_model_name_from_path(file_path)
    data['model_type'] = model_type
    if not unit:
        print('WARNING: No unit is given in write_frame_json: assuming meter.')
        unit = 'meter'
    else:
        assert unit in LENGTH_SCALE_CONVERSION, 'length unit not supported! please use {}'.format(LENGTH_SCALE_CONVERSION.keys())
    data['unit'] = unit
    data['generate_time'] = str(datetime.datetime.now())
    data['dimension'] = len(nodes[0])
    data['node_num'] = len(nodes)
    data['element_num'] = len(elements)
    data['uniform_cross_section'] = unif_cross_sec
    data['uniform_material_properties'] = unif_material
    if unif_cross_sec and unif_material:
        data['material_properties'] = material_dicts[0] if isinstance(material_dicts, list) else material_dicts
        if check_material: assert(check_material_dict(data['material_properties']))
    else:
        data['material_properties'] = {}
        if check_material:
            for mat_dict in material_dicts:
                assert(check_material_dict(mat_dict))
        assert(len(material_dicts) == len(elements))

    data['node_list'] = []
    for i, node in enumerate(nodes):
        assert len(node) == data['dimension'], 'node coordinate not in the same dimension!'
        node_data = OrderedDict()
        node_data['point'] = OrderedDict()
        node_data['point']['X'] = node[0] * LENGTH_SCALE_CONVERSION[unit]
        node_data['point']['Y'] = node[1] * LENGTH_SCALE_CONVERSION[unit]
        if data['dimension'] == 3:
            node_data['point']['Z'] = node[2] * LENGTH_SCALE_CONVERSION[unit]
        node_data['node_id'] = i
        node_data['is_grounded'] = i in fixity_specs
        if fixity_specs:
            node_data['fixities'] = fixity_specs[i] if node_data['is_grounded'] else []
        else:
            node_data['fixities'] = [1] * 6 if node_data['is_grounded'] else []
        data['node_list'].append(node_data)

    data['element_list'] = []
    for i, element in enumerate(elements):
        element_data = OrderedDict()
        element_data['end_node_ids'] = list([int(v) for v in element])
        element_data['element_id'] = i
        element_data['material_properties'] = {} if unif_cross_sec and unif_material else material_dicts[i]
        data['element_list'].append(element_data)

    with open(file_path, 'w+') as outfile:
        if indent:
            json.dump(data, outfile, indent=indent)
        else:
            json.dump(data, outfile)


def read_load_case_json(file_path):
    """Read load case from a json file.

    Note: 
    - For now, only support `kN` for force unit, `kN-m` for moment unit
    - Gravity is assumed to be along the negative global z axis.
    - Element uniform load is converted to global coordinate in this function
    
    Parameters
    ----------
    file_path : str
    
    Returns
    -------
    point_load : dict 
        {node_id : [Fx, Fy, Fz, Mxx, Myy, Mzz]}, in global coordinate
    uniform_element_load : dict 
        {node_id : [wx, wy, wz]}, in global coordinate
    include_self_weight : bool 
        include self-weight or not, now only supports gravity in global z direction
    """
    assert os.path.exists(file_path) and "json file path does not exist!"
    with open(file_path, 'r') as f:
        json_data = json.loads(f.read())

    # version specific sanity checks
    assert json_data['dimension'] == 3, 'model dimension != 3 not supported now!'
    # TODO: do unit conversion here
    assert json_data['force_unit'] == 'kN'
    assert json_data['moment_unit'] == 'kN-m'

    point_load = {}
    uniform_element_load = {}
    include_self_weight = json_data['include_self_weight'] if 'include_self_weight' in json_data else False
    for pl_data in json_data['point_load_list']:
        point_load[pl_data['applied_node_id']] = [pl_data['Fx'], pl_data['Fy'], pl_data['Fz'], 
                                                 pl_data['Mx'], pl_data['My'], pl_data['Mz']]

    if 'uniformly_distributed_element_load_list' in json_data:
        for el_data in json_data['uniformly_distributed_element_load_list']:
            assert el_data['description_frame'] == 'global', 'only description in the global frame supported now!'
            uniform_element_load[el_data['applied_element_id']] = [el_data['wx'], el_data['wy'], el_data['wz']]
    
    return point_load, uniform_element_load, include_self_weight


"""
Some commonly used material properties.

"""
__all__ = ['MATERIAL_PROPERTY_DATABASE']

"""material properties library: Steel, PLA
"""
MATERIAL_PROPERTY_DATABASE = {
    'Steel-S235' : {
        "material_name": "Steel-S235",
        "youngs_modulus": 21000.0,
        "youngs_modulus_unit": "kN/cm2",
        "shear_modulus": 8076.0,
        "shear_modulus_unit": "kN/cm2",
        "density": 78.5,
        "density_unit": "kN/m3",
        "poisson_ratio": 0.300149,
        "cross_sec_area": 19.634954084936208,
        "cross_sec_area_unit": "centimeter^2",
        "_cross_sec_area_note": "round_shape, radius 2.5 cm",
        "Jx": 61.35923151542565,
        "Jx_unit": "centimeter^4",
        "_Jx_note": "0.5 * pi * r^4",
        "Iy": 30.679615757712824,
        "Iy_unit": "centimeter^4",
        "_Iy_note": "0.25 * pi * r^4",
        "Iz": 30.679615757712824,
        "Iz_unit": "centimeter^4",
        "_Iz_note": "0.25 * pi * r^4"
     }, 
    'PLA' : {
        "_source": "https://www.makeitfrom.com/material-properties/Polylactic-Acid-PLA-Polylactide",
        "material_name": "PLA",
        "youngs_modulus": 350,
        "youngs_modulus_unit": "kN/cm2",
        "shear_modulus": 240,
        "shear_modulus_unit": "kN/cm2",
        "tensile_yeild_stress": 3.6,
        "tensile_yeild_stress_unit": "kN/cm2",
        "density": 12.2582,
        "density_unit": "kN/m3",
        "poisson_ratio": -0.27083,
        "cross_sec_area": 0.07068583470577035,
        "cross_sec_area_unit": "centimeter^2",
        "_cross_sec_area_note": "round_shape, radius 0.15 cm",
        "Jx": 0.0007952156404399163,
        "Jx_unit": "centimeter^4",
        "_Jx_note": "0.5 * pi * r^4",
        "Iy": 0.00039760782021995816,
        "Iy_unit": "centimeter^4",
        "_Iy_note": "0.25 * pi * r^4",
        "Iz": 0.00039760782021995816,
        "Iz_unit": "centimeter^4",
        "_Iz_note": "0.25 * pi * r^4"
  }, 
}


# https://github.com/jpanetta/MeshFEM/blob/master/python/init_template.py
################################################################################
#  auto-generated from /Users/yijiangh/conmech/src/pyconmech/init_template.py
################################################################################
import sys as _sys
_sys.path.insert(0, '/Users/yijiangh/conmech/src/pyconmech')

from _pystiffness_checker import _StiffnessChecker



"""
********************************************************************************
pyconmech.database
********************************************************************************

.. currentmodule:: pyconmech.database

This package contains some commonly used material properties, cross section properties
for conmech.

"""

from __future__ import absolute_import

from .material_properties import *


__title__ = 'pyconmech'
__description__ = 'py bindings for conmech: a C++ lib for 3D frame elastis deformation analysis.'
__url__ = 'https://github.com/yijiangh/conmech'
__version__ = '0.4.3'
__author__ = 'Yijiang Huang'
__author_email__ = 'yijiangh@mit.edu'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2019 Yijiang Huang'


"""
********************************************************************************
pyconmech
********************************************************************************

.. currentmodule:: pyconmech

This library provides python wrappers for efficient evaluation of construction mechanics.

.. toctree::
    :maxdepth: 3

    pyconmech.frame_analysis
    pyconmech.database

"""


from __future__ import print_function

import os
import sys
import decimal

# from .cpp_stiffness_checker import *
from .frame_analysis import StiffnessChecker
from .database import *

from .__version__ import __author__, __author_email__, __copyright__, __description__, __license__, __title__, __url__, __version__


__all__ = [
    '__author__', '__author_email__', '__copyright__', '__description__', 
    '__license__', '__title__', '__url__', '__version__', 
    'raise_if_windows',
    'raise_if_not_windows',
    'raise_if_ironpython',
    'raise_if_not_ironpython',
]


def is_windows():
    """Check if the operating system is Windows.
    Returns
    -------
    bool
        True if the OS is Windows. False otherwise
    """
    return os.name == 'nt'
WINDOWS = is_windows()


def is_ironpython():
    """Check if the Python implementation is IronPython.
    Returns
    -------
    bool
        True if the implementation is IronPython. False otherwise
    """
    return 'ironpython' in sys.version.lower()
IPY = is_ironpython()


def raise_if_not_windows():
    if not WINDOWS:
        raise


def raise_if_windows():
    if WINDOWS:
        raise


def raise_if_not_ironpython():
    if not IPY:
        raise


def raise_if_ironpython():
    if IPY:
        raise


"""
********************************************************************************
pyconmech.frame_analysis
********************************************************************************

.. currentmodule:: pyconmech.frame_analysis

Package with functionality to compute elastic deformation of frame structures.

StiffnessChecker
-----------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    StiffnessChecker


File IO
-------

Parsing / saving frame structures from json files 

.. autosummary::
    :toctree: generated/
    :nosignatures:

    read_frame_json
    write_frame_json
    read_load_case_json

Result IO
----------

Parsing / saving analsis results from json files 

.. autosummary::
    :toctree: generated/
    :nosignatures:

    read_frame_analysis_result_json

"""

from __future__ import absolute_import

from .stiffness_checker import *
from .frame_file_io import *
from .result_compare_utils import *
from .stiffness_solve_fn import *

__all__ = [name for name in dir() if not name.startswith('_')]


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from __future__ import absolute_import, print_function

import os
import re
import io

from setuptools import setup, find_packages
from setuptools.command.build_ext import build_ext
from setup_cmake_utils import CMakeExtension, CMakeBuild


here = os.path.abspath(os.path.dirname(__file__))


def read(*names, **kwargs):
    return io.open(
        os.path.join(here, *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()


about = {}
exec(read('src', 'pyconmech', '__version__.py'), about)

requirements = read('requirements.txt').split('\n')

ext_modules = [
    CMakeExtension('_pystiffness_checker'),
    ]


setup(
    name=about['__title__'],
    version=about['__version__'],
    license=about['__license__'],
    description=about['__description__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M |
                   re.S).sub('', read('README.rst')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.rst'))
    ),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={'pyconmech': ['data/*.json']},
    ext_modules=ext_modules,
    cmdclass=dict(build_ext=CMakeBuild),
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        "License :: OSI Approved :: MIT License",
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=['3D frame analysis', 'Finite Element Analysis', 'Structural Analysis'],
    install_requires=requirements,
    # extras_require={},
    # entry_points={},
)
